﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Net.Sockets;
using System.Net;
using System.IO;
using System.IO.Compression;
using System.Drawing.Imaging;

namespace Remote
{
    public partial class Form1 : Form
    {
        private RemoteClient _client;
        public Form1()
        {
            InitializeComponent();
        }
        private void Picture_Received(object sender, PictureEventArgs e)
        {
            pbScreen.Invoke(new MethodInvoker(() =>
                {
                    if (pbScreen.Image != null)
                        pbScreen.Image.Dispose();
                    pbScreen.Image = byteArrayToImage(e.PictureData);
                }));
        }
        private async void Sending_Done(object sender, EventArgs e)
        {
            await Task.Delay(50); // ca. 20 FPS
            _client.Send(Zip.Compress(ImageToByte(takeScreenShot())), 10, tbIP.Text, (int)numPort.Value);
        }
        private Bitmap takeScreenShot()
        {
            var bmpScreenshot = new Bitmap(Screen.PrimaryScreen.Bounds.Width,
                               Screen.PrimaryScreen.Bounds.Height,
                               PixelFormat.Format32bppArgb);
            var gfxScreenshot = Graphics.FromImage(bmpScreenshot);
            gfxScreenshot.CopyFromScreen(Screen.PrimaryScreen.Bounds.X,
                            Screen.PrimaryScreen.Bounds.Y,
                            0,
                            0,
                            Screen.PrimaryScreen.Bounds.Size,
                            CopyPixelOperation.SourceCopy);
            gfxScreenshot.Dispose();
            return bmpScreenshot;
        }
        private ImageCodecInfo GetEncoder(ImageFormat format)
        {
            ImageCodecInfo[] codecs = ImageCodecInfo.GetImageDecoders();

            foreach (ImageCodecInfo codec in codecs)
            {
                if (codec.FormatID == format.Guid)
                {
                    return codec;
                }
            }
            return null;
        }
        public byte[] ImageToByte(Bitmap img)
        {
            byte[] byteArray = new byte[0];
            using (MemoryStream stream = new MemoryStream())
            {
                ImageCodecInfo jgpEncoder = GetEncoder(ImageFormat.Jpeg);

                System.Drawing.Imaging.Encoder myEncoder =
                    System.Drawing.Imaging.Encoder.Quality;

                EncoderParameters myEncoderParameters = new EncoderParameters(1);

                EncoderParameter myEncoderParameter = new EncoderParameter(myEncoder, 5L); // 5L = Qualität (höher = bessere Quali)
                myEncoderParameters.Param[0] = myEncoderParameter;

                img.Save(stream, jgpEncoder, myEncoderParameters);
                stream.Close();

                byteArray = stream.ToArray();
                myEncoderParameters.Dispose();
                myEncoderParameter.Dispose();
            }
            return byteArray;
        }
        public Image byteArrayToImage(byte[] byteArrayIn)
        {
            MemoryStream ms = new MemoryStream(byteArrayIn);
            Image returnImage = Image.FromStream(ms);   
            return returnImage;
        }
        private void btnStart_Click(object sender, EventArgs e)
        {
            if (_client != null)
            _client.StartStream(RemoteClient.Mode.Receiver);
        }
        private void btnStop_Click(object sender, EventArgs e)
        {
            _client.StopStream();
        }
        private void checkBox1_CheckedChanged(object sender, EventArgs e)
        {
            if (cbTransfer.Checked)
            {
                _client.StartStream(RemoteClient.Mode.Sender);
                _client.Send(Zip.Compress(ImageToByte(takeScreenShot())), 10, tbIP.Text, (int)numPort.Value);
            }
            else
                _client.StopStream();
        }
        private void btnBind_Click(object sender, EventArgs e)
        {
            _client = new RemoteClient((int)numPort.Value);
            _client.PictureReceived += new EventHandler<PictureEventArgs>(Picture_Received);
            _client.SendingDone += new EventHandler<EventArgs>(Sending_Done);
            btnBind.Enabled = false;
            cbTransfer.Enabled = true;
            btnStart.Enabled = true;
            btnStop.Enabled = true;
            tbIP.Enabled = true;
        }
    }
    
    public class RemoteClient
    {
        private bool _running;
        private UdpClient _client;
        public RemoteClient()
        {
            _client = new UdpClient(8000);
        }
        public RemoteClient(int pPort)
        {
            _client = new UdpClient(pPort);
        }
        public void Send(byte[] pBytes, string pIP, int pPort)
        {
            _client.Send(pBytes, pBytes.Length, pIP, pPort);
        }
        private void EndReceive(IAsyncResult pResult)
        {
            IPEndPoint lEndPoint = new IPEndPoint(IPAddress.Any, 0);
            byte[] lBytes = _client.EndReceive(pResult, ref lEndPoint);
            if (lBytes[0] == 1)
                SendingDone(this, EventArgs.Empty);
            if (_running)
            _client.BeginReceive(EndReceive, null);
        }
        public void StartStream(Mode pMode)
        {
            if (_running)
                return;
            _running = true;
            if (pMode == Mode.Receiver)
                _client.BeginReceive(EndReceiveData, null);
            else if (pMode == Mode.Sender)
                _client.BeginReceive(EndReceive, null);
            _mode = pMode;
        }
        public void StopStream()
        {
           _running = false;
        }
        private Mode _mode;
        public enum Mode
        {
            Sender,
            Receiver
        }
        public void Send(byte[] pBytes, int pCount, string pIP, int pPort)
        {
            int lPartLength = (int)Math.Floor((double)(pBytes.Length / pCount));
            for (int i = 0; i < pCount + 1; i++)
            {
                List<byte> lBytes = pBytes.Skip(i * lPartLength).ToList();
                lBytes.Insert(0, (byte)(i + 1));
                byte[] lSendingBytes;
                if (i * lPartLength + lPartLength < pBytes.Length)
                    lSendingBytes = lBytes.Take(lPartLength + 1).ToArray();
                else
                    lSendingBytes = lBytes.ToArray();
                byte[] lCompressed = lSendingBytes;
                Send(lCompressed, pIP, pPort);
            }
        }
        public event EventHandler<EventArgs> SendingDone;
        private const int ENDOFPART = 11;
        private byte[][] imgBuffer = new byte[ENDOFPART][];
        private int _received = 0;
        private void EndReceiveData(IAsyncResult pResult)
        {
            lock (this)
            {
                IPEndPoint lEndPoint = new IPEndPoint(IPAddress.Any, 0);
                byte[] lData = _client.EndReceive(pResult, ref lEndPoint);
                    byte lIndex = lData[0];
                    _received++;
                    imgBuffer[lIndex - 1] = lData.Skip(1).ToArray();
                    if (_received == ENDOFPART)
                    {
                        byte[][] lCopy = (byte[][])imgBuffer.Clone();
                        int lLength = lCopy.Sum((item) => item.Length);
                        byte[] allBytes = new byte[lLength];
                        for (int i = 0; i < ENDOFPART; i++)
                            System.Buffer.BlockCopy(lCopy[i], 0, allBytes, i * imgBuffer[0].Length, imgBuffer[i].Length);
                        _received = 0;
                        _client.Send(new byte[] { 1 }, 1, lEndPoint);
                        PictureReceived(this, new PictureEventArgs(Zip.Decompress(allBytes)));
                    }
                if (_running)
                _client.BeginReceive(EndReceiveData, null);
            }
        }
        public event EventHandler<PictureEventArgs> PictureReceived;
        
    }
    public class PictureEventArgs : EventArgs
    {
        private readonly byte[] _picture;
        public PictureEventArgs (byte[] pData)
        {
            this._picture = pData;
        }
        public byte[] PictureData { get { return _picture; } }
    }
    public static class Zip
    {
        public static byte[] Decompress(byte[] zippedData)
        {
            byte[] decompressedData = null;
            using (MemoryStream outputStream = new MemoryStream())
            {
                using (MemoryStream inputStream = new MemoryStream(zippedData))
                {
                    using (GZipStream zip = new GZipStream(inputStream, CompressionMode.Decompress))
                    {
                        zip.CopyTo(outputStream);
                    }
                }
                decompressedData = outputStream.ToArray();
            }

            return decompressedData;
        }
        public static byte[] Compress(byte[] plainData)
        {
            byte[] compressesData = null;
            using (MemoryStream outputStream = new MemoryStream())
            {
                using (GZipStream zip = new GZipStream(outputStream, CompressionMode.Compress))
                {
                    zip.Write(plainData, 0, plainData.Length);
                }
                compressesData = outputStream.ToArray();
            }

            return compressesData;
        }
    }
}
